"""# Skáld

📟 a simple and efficient experiment logger for Python 🐍
"""

from datetime import datetime
from typing import Any, Self
import yaml
from pathlib import Path
from beartype import beartype
from loguru import logger
from loguru._logger import Logger as LoguruLogger
import polars as pl
from polars import DataFrame
import sys
import contextlib

from skald.enums import (
    PersistenceStrategy,
    PersistenceStrategyLit,
    MetricsFileFormat,
    MetricsFileFormatLit,
)
from skald.utils import flatten_dict
from skald import tui


@beartype
class Logger:
    """📃 an experiment logger for storing console logs, metrics, parameters and artifacts.

    This class integrates an optional [textualize](https://textual.textualize.io/) TUI app for viewing the logs.

    ???+ info

        Skáld also provides a CLI tool (`skald`) to view logs later.
        simply call `$skald <path_to_log_dir>` or `$skald --help` for help.

    Attributes:
        dir (Path): directory, where the log "run" directory will be created.
        run_name (str): name of the experiment/run. Can be defined by the user.
            If not defined, the current timestamp will be used.
        run_dir (Path): path to the run directory. Will be `dir/{run_name}`.
        timestamp (datetime): timestamp of the creation of the logger instance.
        metrics_file (Path): path to the metrics file.
        metrics_file_format (MetricsFileFormat): file format of the metrics.
            Currently, csv and parquet are supported. Defaults to "parquet".
        params_file (Path): path to the parameters file.
        artifacts_dir (Path): path to the artifacts directory.
        persistence_strategy (PersistenceStrategy): when to save the metrics to disk.
            Used for performance optimizations. Defaults to "eager".
        tui (bool): Whether to start the experiment viewer terminal user interface.
    """

    # 🔎 Logger and Run Details
    dir: Path
    run_name: str
    run_dir: Path
    timestamp: datetime
    metrics_file: Path
    metrics_file_format: MetricsFileFormat
    params_file: Path
    artifacts_dir: Path
    persistence_strategy: PersistenceStrategy
    tui: bool

    # 🔢 Internals
    _logger: LoguruLogger
    _metrics: DataFrame
    _params: dict[str, Any]

    def __init__(
        self,
        dir: Path = Path("."),
        run_name: str | None = None,
        persistence_strategy: PersistenceStrategy | PersistenceStrategyLit = "eager",
        metrics_file_format: MetricsFileFormat | MetricsFileFormatLit = "parquet",
        tui: bool = False,
    ):
        """🏃‍♂️ creates a logger instance and prepares a Skáld run.

        Each run will have the following structure:

        - `dir/{run_name}/`
            - `console.log`
            - `metrics.{metrics_file_format}`
            - `params.yaml`
            - `artifacts/`

        Args:
            dir (Path, optional): path to the directory under which the run directory will be created. Defaults to Path(".").
            run_name (str | None, optional): name of the run. If None, use the current time as `run_name`. Defaults to None.
            persistence_strategy (PersistenceStrategy | PersistenceStrategyLit, optional): when to save the metrics to disk.
                Used for performance optimizations. Defaults to "eager".
            metrics_file_format (MetricsFileFormat | MetricsFileFormatLit, optional): file format of the metrics. Defaults to "parquet".
            tui (bool, optional): Whether to start the experiment viewer terminal user interface. Defaults to false.
        """
        self.timestamp = datetime.now()
        self.run_name = (
            run_name if run_name else self.timestamp.strftime("%Y%m%d-%H%M%S")
        )
        self.run_dir = dir / self.run_name

        self.run_dir.mkdir()

        # 📃 console logging with loguru
        # * we also redirect stdout to the log file, which requires a trick from:
        # * https://loguru.readthedocs.io/en/stable/resources/recipes.html#capturing-standard-stdout-stderr-and-warnings
        self._logger = logger

        class _StreamToLogger:
            def __init__(self, _logger):
                self._logger = _logger

            def write(self, buffer):
                self._logger.opt(depth=1, raw=True).info(buffer)

            def flush(self):
                pass

        self._logger.remove()
        self._logger.add(sys.__stdout__)
        self._logger.add(f"{self.run_dir/'console.log'}")
        self._redirect_stdout = contextlib.redirect_stdout(
            _StreamToLogger(self._logger)
        )

        self.info(f"📂 logging to {self.run_dir}")

        self.persistence_strategy = persistence_strategy

        # logging run meta data in run_dir / params_file
        self._params = {
            "skald.timestamp": str(self.timestamp),
            "skald.run_name": self.run_name,
            "skald.version": "v1",
        }
        self.params_file = self.run_dir / Path("params.yaml")
        self._save_params()

        self._metrics = DataFrame()
        self.metrics_file = self.run_dir / Path("metrics").with_suffix(
            f".{metrics_file_format}"
        )
        self.metrics_file_format = metrics_file_format
        self._save_metrics()

        self.artifacts_dir = self.run_dir / Path("artifacts")
        self.artifacts_dir.mkdir()

        self.tui = tui

    def log(self):
        raise NotImplementedError

    def get_metrics_with_params(self) -> DataFrame:
        """🐻‍❄️ returns the metrics with params in additional columns.

        Returns:
            DataFrame: tidy metrics + params
        """
        return self._metrics.with_columns(
            [pl.lit(param).alias(name) for param, name in self._params.items()]
        )

    # 📈 Metrics

    def log_metric(self, name: str, value: int | float, **ids: dict) -> None:
        """📈 logs a metric with identifiers given as keyword arguments.

        If `self.persistence_strategy` is `"eager"`, the metric will be saved to disk.

        Args:
            name (str): name of the metric
            value (int | float): value of the metric
            **ids: arbitrary keyword arguments used as metric identifier.
        """
        if isinstance(value, int):
            value = float(value)

        new_entry = DataFrame({"name": name, "value": value} | ids)
        self._metrics = pl.concat([self._metrics, new_entry], how="diagonal")

        match self.persistence_strategy:
            case PersistenceStrategy.EAGER:
                self._save_metrics()

    def log_scalar(self, name: str, value: int | float, **ids: dict) -> None:
        """📈 logs a scalar (metric) with identifiers given as keyword arguments.

        If `self.persistence_strategy` is `"eager"`, the metric will be saved to disk.

        Args:
            name (str): name of the metric
            value (int | float): value of the metric
            **ids: arbitrary keyword arguments used as metric identifier.
        """
        self.log_metric(name, value, **ids)

    def log_metrics(self, metrics: dict[str, int | float], **ids) -> None:
        """📈 logs a dictionary of metrics with identifiers given as keyword arguments.

        If `self.persistence_strategy` is `"eager"`, the metrics will be saved to disk.

        Examples:
            An example of how to log two metrics stored in a dictionary.

            >>> logger.log_metrics({"loss": 1.23, "accuracy": 0.9}, stage="test", step=1)


        Args:
            metrics (dict[str, int | float]): single-level dictionary of metrics.
            **ids: arbitrary keyword arguments used as metric identifier.
        """
        for name, value in metrics.items():
            self.log_scalar(name, value, **ids)

    def log_scalars(self, metrics: dict[str, int | float], **ids) -> None:
        """📈 logs a dictionary of scalars (metrics) with identifiers given as keyword arguments.

        If `self.persistence_strategy` is `"eager"`, the metrics will be saved to disk.

        Examples:
            An example of how to log two metrics stored in a dictionary.

            >>> logger.log_metrics({"loss": 1.23, "accuracy": 0.9}, stage="test", step=1)


        Args:
            metrics (dict[str, int | float]): single-level dictionary of metrics.
            **ids: arbitrary keyword arguments used as metric identifier.
        """
        for name, value in metrics.items():
            self.log_scalar(name, value, **ids)

    def log_dict(self, metrics: dict[str, int | float], **ids) -> None:
        """📈 logs a dictionary of metrics with identifiers given as keyword arguments.

        If `self.persistence_strategy` is `"eager"`, the metrics will be saved to disk.

        Examples:
            An example of how to log two metrics stored in a dictionary.

            >>> logger.log_metrics({"loss": 1.23, "accuracy": 0.9}, stage="test", step=1)


        Args:
            metrics (dict[str, int | float]): single-level dictionary of metrics.
            **ids: arbitrary keyword arguments used as metric identifier.
        """
        for name, value in metrics.items():
            self.log_scalar(name, value, **ids)

    # ⚙️ Parameters

    def log_param(self, name: str, value: Any) -> None:
        """⚙️ logs a single parameter.

        If the parameter exists, it will be overwritten.

        If `self.persistence_strategy` is `"eager"`, the parameter will be saved to disk.

        Args:
            name (str): name of the parameter.
            value (Any): value of the parameter.
                parameters are stored as yaml files, so any yaml-compatible value is allowed.
        """
        self._params[name] = value

        match self.persistence_strategy:
            case PersistenceStrategy.EAGER:
                self._save_params()

    def log_arg(self, name: str, value: Any) -> None:
        """⚙️ logs a single argument (parameter).

        If `self.persistence_strategy` is `"eager"`, the parameter will be saved to disk.

        Args:
            name (str): name of the parameter.
            value (Any): value of the parameter.
                parameters are stored as yaml files, so any yaml-compatible value is allowed.
        """
        self.log_param(name, value)

    def log_params(self, params: dict, separator: str = ".") -> None:
        """⚙️ logs a dictionary of parameters.

        To facilitate the consumption and visualization, the parameters dictionary will be flattened.

        If `self.persistence_strategy` is `"eager"`, the parameters will be saved to disk.

        Args:
            params (dict): (possibly nested) dictionary of parameters.
            separator (str, optional): keys are concatenated using this separator. Defaults to '.'.
        """
        params = flatten_dict(params, sep=separator)

        for param, value in params.items():
            self.log_param(param, value)

    def log_args(self, params: dict, separator: str = ".") -> None:
        """⚙️ logs a dictionary of arguments (parameters).

        To facilitate the consumption and visualization, the parameters dictionary will be flattened.

        If `self.persistence_strategy` is `"eager"`, the parameters will be saved to disk.

        Args:
            params (dict): (possibly nested) dictionary of parameters.
            separator (str, optional): keys are concatenated using this separator. Defaults to '.'.
        """
        self.log_params(params, separator)

    # 🖼️ Artifacts

    def log_figure(self):
        raise NotImplementedError

    def log_image(self):
        raise NotImplementedError

    # 💾 Saving to Disk

    def save(self) -> None:
        """💾 saves metrics and parameters to disk."""
        match self.persistence_strategy:
            case PersistenceStrategy.EAGER:
                self.info(
                    "There is no need to call `save` when using `PersistenceStrategy.EAGER`."
                )

        self._save_metrics()
        self._save_params()

    def _save_metrics(self) -> None:
        """💾 saves metrics to disk."""
        match self.metrics_file_format:
            case "csv":
                self._metrics.write_csv(self.metrics_file)
            case "parquet":
                self._metrics.write_parquet(self.metrics_file)

    def _save_params(self) -> None:
        """💾 saves parameters to disk."""
        with (self.params_file).open("w") as f:
            yaml.dump(self._params, f)

    def __enter__(self) -> Self:
        """📲 called when entering a context manager."""
        self._redirect_stdout.__enter__()
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        """🚪 called when leaving a context manager.

        The parameters and metrics will be saved.
        """
        self.save()
        self._redirect_stdout.__exit__(exc_type, exc_value, traceback)
        # * because the tui is not yet fully featured (metrics are not updating),
        # * we only start the tui after the log has finished to inspect the logs.
        if self.tui:
            tui.view_experiment(self.run_dir)

    # 📃 Console Logging

    def debug(self, message: str, *args, **kwargs) -> None:
        """🔎 logs a debug message using [loguru](https://loguru.readthedocs.io/en/stable/)."""
        self._logger.debug(message, *args, **kwargs)

    def info(self, message: str, *args, **kwargs) -> None:
        """ℹ️ logs an info message using [loguru](https://loguru.readthedocs.io/en/stable/)."""
        self._logger.info(message, *args, **kwargs)

    def warning(self, message: str, *args, **kwargs) -> None:
        """❗ logs a warning message using [loguru](https://loguru.readthedocs.io/en/stable/)."""
        self._logger.warning(message, *args, **kwargs)

    def error(self, message: str, *args, **kwargs) -> None:
        """🔥 logs n error message using [loguru](https://loguru.readthedocs.io/en/stable/)."""
        self._logger.error(message, *args, **kwargs)

    def critical(self, message: str, *args, **kwargs) -> None:
        """🤯 logs a critical message using [loguru](https://loguru.readthedocs.io/en/stable/)."""
        self._logger.critical(message, *args, **kwargs)

    def success(self, message: str, *args, **kwargs) -> None:
        """✅ logs a success message using [loguru](https://loguru.readthedocs.io/en/stable/)."""
        self._logger.success(message, *args, **kwargs)



"""🧑‍💻 terminal user interfaces"""

from pathlib import Path
from typing import Optional

import polars as pl
import typer
from typing_extensions import Annotated
from textual.app import App, ComposeResult
from textual.widgets import (
    Header,
    TabbedContent,
    TabPane,
    RichLog,
    Pretty,
    DirectoryTree,
    Static,
)
from textual_plotext import PlotextPlot
import yaml
from aiofile import async_open

from textual.containers import ScrollableContainer


class LogsWidget(RichLog):
    """📃 a widget, that periodically reads `logs_file` and displays the result."""

    def __init__(self, logs_file: Path, update_interval: float | None = None):
        self.logs_file = logs_file
        self.update_interval = update_interval
        super().__init__()

    async def on_mount(self) -> None:
        await self.read_logs()
        if self.update_interval:
            self.update_timer = self.set_interval(self.update_interval, self.read_logs)

    async def read_logs(self) -> None:
        async with async_open(self.logs_file, "r") as f:
            logs = await f.read()

        self.clear()
        self.write(logs)


class MetricsWidget(PlotextPlot):
    """📉 a static widget, that plots a metric using [plottext](https://github.com/Textualize/textual-plotext)."""

    def __init__(self, name: str, values: list[float], ids: list, id: str):
        self.metric_name = name
        self.metric_values = values
        self.metric_ids = ids
        super().__init__(id=id)

    def on_mount(self) -> None:
        self.plt.plot(self.metric_values, self.metric_ids)
        self.plt.title(self.metric_name)


class StaticMetricsListWidget(Static):
    """🔎📈 a static widget, that reads a `metrics_file` and plots each metric."""

    def __init__(self, metrics_file: Path):
        self.metrics_file = metrics_file
        super().__init__()

    def compose(self) -> ComposeResult:
        """Create child widgets for the app."""
        self.read_metrics()

        plots = self.get_plots()

        yield ScrollableContainer(*plots, id="plots")

    def get_plots(self) -> list[MetricsWidget]:
        plots = []
        # parse metrics and decide what to plot
        for (name,), df in self.metrics.groupby(["name"]):
            # get values
            values = df["value"].to_list()
            # get ids
            # * in this first prototype, we just arange the values
            ids = range(len(values))

            plots.append(MetricsWidget(name, values, ids, id=f"{name}-plot"))

        return plots

    async def read_metrics_async(self) -> None:
        async with async_open(self.metrics_file, "rb") as f:
            metrics = await f.read()
        match self.metrics_file.suffix:
            case ".csv":
                self.metrics = pl.read_csv(metrics)
            case ".parquet":
                self.metrics = pl.read_parquet(metrics)

    def read_metrics(self) -> None:
        match self.metrics_file.suffix:
            case ".csv":
                self.metrics = pl.read_csv(self.metrics_file)
            case ".parquet":
                self.metrics = pl.read_parquet(self.metrics_file)


class ParametersWidget(Pretty):
    """⚙️ a widget, that periodically reads `params_file` and displays the result."""

    def __init__(self, params_file: Path, update_interval: float | None = None):
        self.params_file = params_file
        self.update_interval = update_interval
        super().__init__(None)

    async def on_mount(self) -> None:
        await self.read_params()
        if self.update_interval:
            self.update_timer = self.set_interval(
                self.update_interval, self.read_params
            )

    async def read_params(self) -> None:
        async with async_open(self.params_file, "r") as f:
            file_content = await f.read()
        params = yaml.safe_load(file_content)
        self.update(params)


class ArtifactsWidget(DirectoryTree):
    """📂 a simple directory tree widget that updates periodically."""

    def __init__(self, artifacts_dir: Path, update_interval: float | None = None):
        self.artifacts_dir = artifacts_dir
        self.update_interval = update_interval
        super().__init__(artifacts_dir)

    def on_mount(self) -> None:
        if self.update_interval:
            self.update_timer = self.set_interval(self.update_interval, self.reload)


class ExperimentViewerApp(App):
    """🔎⚗️ a Textual app to view Skáld experiments."""

    TITLE = "🎶 Skáld"
    SUB_TITLE = "an experiment viewer"

    BINDINGS = [
        ("l", "switch_tab('tab-1')", "📃 view logs"),
        ("m", "switch_tab('tab-2')", "📈 view metrics"),
        ("p", "switch_tab('tab-3')", "⚙️ view parameters"),
        ("a", "switch_tab('tab-4')", "📂 view artifacts"),
        ("d", "toggle_dark", "🌒 toggle dark mode"),
    ]

    CSS_PATH = "tui.tcss"

    def __init__(self, experiment_dir: Path, update_interval: float | None = None):
        """Creates an experiment viewer for a Skáld experiment in `experiment_dir`.

        Args:
            experiment_dir (Path): directory containing logs of a Skáld run
            update_interval (float | None, optional): time in seconds to update the widgets
                by reading the files again. Defaults to None
        """
        self.experiment_dir = experiment_dir
        self.update_interval = update_interval
        super().__init__()

    def compose(self) -> ComposeResult:
        """Create child widgets for the app."""
        yield Header(show_clock=True)
        with TabbedContent():
            with TabPane("📃 logs"):
                yield LogsWidget(
                    self.experiment_dir / "console.log", self.update_interval
                )
            with TabPane("📈 metrics"):
                yield StaticMetricsListWidget(
                    next(self.experiment_dir.glob("metrics.*"))
                )
            with TabPane("⚙️ params"):
                yield ParametersWidget(
                    self.experiment_dir / "params.yaml", self.update_interval
                )
            with TabPane("📂 artifacts"):
                yield ArtifactsWidget(
                    self.experiment_dir / "artifacts", self.update_interval
                )
        # yield Footer() #* the footer seems buggy in dark mode -> duplicated footers, missing lines and scrolling issues...

    def action_switch_tab(self, tab: str) -> None:
        """Change the active tab.

        Args:
            tab (str): new active tab
        """
        self.query_one(TabbedContent).active = tab

    def action_toggle_dark(self) -> None:
        """An action to toggle dark mode."""
        self.dark = not self.dark


def view_experiment(
    experiment_dir: Path,
    update_interval: Annotated[Optional[float], typer.Option()] = None,
):
    """🔎⚗️ view a Skáld experiment in `experiment_dir` with a [textualize](https://textual.textualize.io/) app.

    Args:
        experiment_dir (Path): directory containing logs of a Skáld run
        update_interval (float | None, optional): time in seconds to update the widgets
            by reading the files again. Defaults to None
    """
    app = ExperimentViewerApp(experiment_dir, update_interval)
    app.run()


def main():
    typer.run(view_experiment)


if __name__ == "__main__":
    main()



"""🛠️ utility functions for `skald`."""

import pandas as pd


def flatten_dict(d: dict, sep: str = ".") -> dict:
    """📖 flattens a nested dictionary with only keys of type `str` into a single-level dictionary by concatenating keys with <sep>.

    Code from https://www.freecodecamp.org/news/how-to-flatten-a-dictionary-in-python-in-4-different-ways/

    Args:
        d (dict): dictionary
        sep (str, optional): keys are concatenated using this separator. Defaults to '.'.

    Returns:
        dict: flattened dictionary
    """
    [flat_dict] = pd.json_normalize(d, sep=sep).to_dict(orient="records")
    return flat_dict
